import argparse
import pathlib
import sys
from lark import __version__ as lark_version
from gersemi.configuration import (
    make_configuration,
    make_default_configuration_file,
    Configuration,
)
from gersemi.mode import get_mode
from gersemi.return_codes import SUCCESS, FAIL
from gersemi.runner import run, print_to_stderr
from gersemi.__version__ import __title__, __version__


class GenerateConfigurationFile(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print(make_default_configuration_file())


class ShowVersion(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print(f"{__title__} {__version__}")
        print(f"lark {lark_version}")
        print(f"Python {sys.version}")


def create_argparser():
    parser = argparse.ArgumentParser(
        description="A formatter to make your CMake code the real treasure.",
        prog="gersemi",
        add_help=False,
    )
    modes_group = parser.add_argument_group("modes")
    modes_group.add_argument(
        "-c",
        "--check",
        dest="check_formatting",
        action="store_true",
        help=f"Check if files require reformatting. "
        f"Return {SUCCESS} when there's nothing to reformat, "
        f"return {FAIL} when some files would be reformatted",
    )
    modes_group.add_argument(
        "-i",
        "--in-place",
        dest="in_place",
        action="store_true",
        help="Format files in-place",
    )
    modes_group.add_argument(
        "--diff",
        dest="show_diff",
        action="store_true",
        help="Show diff on stdout for each formatted file instead",
    )
    modes_group.add_argument(
        "--default-config",
        nargs=0,
        action=GenerateConfigurationFile,
        help="Generate default .gersemirc configuration file",
    )
    modes_group.add_argument(
        "--version",
        nargs=0,
        dest="show_version",
        action=ShowVersion,
        help="Show version.",
    )
    modes_group.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )

    configuration_group = parser.add_argument_group(
        title="configuration",
        description="By default configuration is loaded from YAML formatted .gersemirc "
        "file if it's available. "
        "This file should be placed in one of the common parent directories of source files. "
        "Arguments from command line can be used to override parts of that configuration "
        "or supply them in absence of configuration file.",
    )
    configuration_group.add_argument(
        "-l",
        "--line-length",
        metavar="INTEGER",
        dest="line_length",
        type=int,
        help=f"Maximum line length in characters [default: {Configuration().line_length}]",
    )
    configuration_group.add_argument(
        "--unsafe",
        dest="unsafe",
        action="store_true",
        help="Skip default sanity checks",
    )
    configuration_group.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="Skip printing non-error messages to stderr",
    )
    configuration_group.add_argument(
        "--color",
        dest="color",
        action="store_true",
        help="If --diff is selected showed diff is colorized",
    )
    configuration_group.add_argument(
        "--definitions",
        dest="definitions",
        metavar="src",
        default=[],
        nargs="+",
        type=pathlib.Path,
        help="Files or directories containing custom command definitions (functions or macros). "
        "If only - is provided custom definitions, if there are any, are taken from stdin instead. "
        "Commands from not deprecated CMake native modules don't have to be provided "
        "(check https://cmake.org/cmake/help/latest/manual/cmake-modules.7.html)",
    )
    configuration_group.add_argument(
        "--list-expansion",
        dest="list_expansion",
        choices=["favour-inlining", "favour-expansion"],
        help="Switch controls how code is expanded into multiple lines when it's not possible "
        "to keep it formatted in one line. With 'favour-inlining' (default) the list of entities "
        "will be formatted in such way that sublists might still be formatted into single line "
        "as long as it's possible. With 'favour-expansion' the list of entities will be formatted "
        "in such way that sublists will be completely expanded once expansion becomes necessary "
        "at all.",
    )

    parser.add_argument(
        dest="sources",
        metavar="src",
        nargs="*",
        type=pathlib.Path,
        help="File or directory to format. "
        "If only - is provided input is taken from stdin instead",
    )

    return parser


def is_stdin_mixed_with_file_input(sources):
    return pathlib.Path("-") in sources and len(sources) != 1


def postprocess_args(args):
    args.sources = set(args.sources)
    args.definitions = set(args.definitions)


def main():
    argparser = create_argparser()
    args = argparser.parse_args()
    postprocess_args(args)

    if any(map(is_stdin_mixed_with_file_input, [args.sources, args.definitions])):
        print_to_stderr("Don't mix stdin with file input")
        sys.exit(FAIL)

    if len(args.sources) < 1:
        sys.exit(SUCCESS)

    configuration = make_configuration(args)
    mode = get_mode(args)

    sys.exit(run(mode, configuration, args.sources))


if __name__ == "__main__":
    main()



from typing import Sequence
from lark import Tree
from gersemi.command_invocation_dumpers.argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)
from gersemi.keywords import Keywords


def create_specialized_dumper(positional_arguments: Sequence[str], keywords: Keywords):
    class Impl(ArgumentAwareCommandInvocationDumper):
        front_positional_arguments = positional_arguments
        options = keywords.options
        one_value_keywords = keywords.one_value_keywords
        multi_value_keywords = keywords.multi_value_keywords

        def custom_command(self, tree):
            _, command_name, arguments, *_ = tree.children
            return self.visit(Tree("command_invocation", [command_name, arguments]))

    return Impl



# pylint: disable=unnecessary-lambda-assignment
from contextlib import contextmanager, ExitStack
import filecmp
import os
from pathlib import Path
import re
import shutil
from stat import S_IREAD, S_IRGRP, S_IROTH
import subprocess
import sys
import tempfile
import yaml
from tests.cache_inspector import inspect_cache


def gersemi_with_cache_path(cache_path, *gersemi_args, **subprocess_kwargs):
    this_file_dir = os.path.dirname(os.path.realpath(__file__))
    subprocess_kwargs["input"] = cache_path + "\n" + subprocess_kwargs.get("input", "")
    return subprocess.run(
        [sys.executable, f"{this_file_dir}/patched_gersemi", *gersemi_args],
        check=False,
        encoding="utf8",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **subprocess_kwargs,
    )


@contextmanager
def make_temporary_cache():
    with tempfile.TemporaryDirectory() as directory:
        temporary_cache = str(Path(directory) / "temporary_cache.db")
        yield temporary_cache


def gersemi(*gersemi_args, **subprocess_kwargs):
    with make_temporary_cache() as temporary_cache:
        return gersemi_with_cache_path(
            temporary_cache, *gersemi_args, **subprocess_kwargs
        )


HERE = os.path.dirname(os.path.realpath(__file__))


def case(filepath):
    return f"{HERE}/executable/{filepath}"


@contextmanager
def temporary_copy(original):
    with tempfile.TemporaryDirectory() as temp_directory:
        original_base = os.path.basename(original)
        temp_path = os.path.join(temp_directory, original_base)
        try:
            shutil.copy2(original, temp_path)
            yield temp_path
        finally:
            os.remove(temp_path)


@contextmanager
def temporary_copies(file_paths):
    with ExitStack() as stack:
        copies = [stack.enter_context(temporary_copy(p)) for p in file_paths]
        yield copies


@contextmanager
def temporary_dir_copy(original):
    with tempfile.TemporaryDirectory() as temp_directory:
        original_base = os.path.basename(original)
        temp_path = os.path.join(temp_directory, original_base)
        try:
            shutil.copytree(original, temp_path)
            yield temp_path
        finally:
            shutil.rmtree(temp_path)


@contextmanager
def temporary_dir_copies(directory_paths):
    with ExitStack() as stack:
        copies = [stack.enter_context(temporary_dir_copy(p)) for p in directory_paths]
        yield copies


def assert_success(*args, **kwargs):
    process = gersemi(*args, **kwargs)
    assert process.returncode == 0, (process.stdout, process.stderr)


def assert_fail(*args, **kwargs):
    assert gersemi(*args, **kwargs).returncode == 1


def compare_directories(left, right):
    comparison = filecmp.dircmp(left, right)
    return {
        "left_only": comparison.left_only,
        "right_only": comparison.right_only,
        "diff_files": comparison.diff_files,
        "funny_files": comparison.funny_files,
    }


@contextmanager
def create_dot_gersemirc(where, **kwargs):
    p = os.path.join(where, ".gersemirc")
    try:
        with open(p, "w", encoding="utf-8") as f:
            f.write(yaml.dump(kwargs))
        yield
    finally:
        os.remove(p)


@contextmanager
def create_fake_definitions(where, name):
    p = os.path.join(where, name)
    try:
        with open(os.path.join(where, name), "w", encoding="utf-8") as f:
            f.write("\n")
        yield p
    finally:
        os.remove(p)


@contextmanager
def create_configuration_files(directories, creator):
    with ExitStack() as stack:
        configuration_files = [
            stack.enter_context(create_dot_gersemirc(where=d, **creator(d)))
            for d in directories
        ]
        yield configuration_files


def assert_that_directories_differ(left, right):
    comparison = compare_directories(left, right)
    for value in comparison.values():
        if len(value) > 0:
            return
    raise AssertionError("directories have the same content")


def assert_that_directories_have_the_same_content(left, right):
    comparison = compare_directories(left, right)
    report = []
    for key, value in comparison.items():
        if len(value) > 0:
            report += [f"{key}: {value}"]

    assert len(report) == 0, "directories differences:\n" + "\n".join(report)


def test_help_just_works():
    assert_success("--help")


def test_check_on_formatted_file_should_return_zero():
    with temporary_copy(case("formatted_file.cmake")) as copy:
        assert_success("--check", copy)


def test_check_on_not_formatted_file_should_return_one():
    with temporary_copy(case("not_formatted_file.cmake")) as copy:
        assert_fail("--check", copy)


def test_format_file_in_place():
    with temporary_copy(case("not_formatted_file.cmake")) as copy:
        assert_fail("--check", copy)
        gersemi("--in-place", copy)
        assert_success("--check", copy)


def test_check_formatted_input_from_stdin():
    inp = """set(FOO BAR)
"""
    assert_success("--check", "-", input=inp)


def test_check_not_formatted_input_from_stdin():
    inp = """set(FOO BAR)"""  # missing newline at the end
    assert_fail("--check", "-", input=inp)


def test_format_formatted_input_from_stdin():
    inp = """set(FOO BAR)
"""
    completed_process = gersemi("-", input=inp)
    assert completed_process.returncode == 0
    assert completed_process.stdout == inp
    assert completed_process.stderr == ""


def test_format_not_formatted_input_from_stdin():
    inp = "set(FOO BAR)"  # missing newline at the end

    completed_process = gersemi("-", input=inp)
    assert completed_process.returncode == 0
    assert completed_process.stdout == inp + "\n"
    assert completed_process.stderr == ""


def test_when_run_with_no_input_should_return_zero():
    assert_success()  # no args


def test_dont_mix_stdin_and_file_input():
    assert_fail(case("formatted_file.cmake"), "-")
    assert_fail("-", case("formatted_file.cmake"))
    assert_fail(case("formatted_file.cmake"), "-", case("formatted_file.cmake"))


def test_check_multiple_formatted_input_files():
    case_ = lambda filename: case("/directory_with_formatted_files/" + filename)
    with temporary_copies(
        [case_("file1.cmake"), case_("file2.cmake"), case_("file3.cmake")]
    ) as copies:
        assert_success("--check", *copies)


def test_check_multiple_not_formatted_input_files():
    case_ = lambda filename: case("/directory_with_not_formatted_files/" + filename)
    with temporary_copies(
        [case_("file1.cmake"), case_("file2.cmake"), case_("file3.cmake")]
    ) as copies:
        assert_fail("--check", *copies)


def test_check_multiple_input_files_when_some_are_not_formatted():
    case_ = lambda filename: case(
        "/directory_with_some_not_formatted_files/" + filename
    )
    with temporary_copies(
        [
            case_("formatted_file1.cmake"),
            case_("formatted_file2.cmake"),
            case_("formatted_file3.cmake"),
            case_("not_formatted_file1.cmake"),
            case_("not_formatted_file2.cmake"),
            case_("not_formatted_file3.cmake"),
        ]
    ) as copies:
        assert_fail("--check", *copies)


def test_check_directory_with_formatted_files():
    with temporary_dir_copy(case("directory_with_formatted_files")) as copy:
        assert_success("--check", copy)


def test_check_directory_with_not_formatted_files():
    with temporary_dir_copy(case("directory_with_not_formatted_files")) as copy:
        assert_fail("--check", copy)


def test_check_directory_with_some_not_formatted_files():
    with temporary_dir_copy(case("directory_with_some_not_formatted_files")) as copy:
        assert_fail("--check", copy)


def test_format_in_place_multiple_formatted_files():
    files = ["file1.cmake", "file2.cmake", "file3.cmake"]
    with temporary_copies(
        case("/directory_with_formatted_files/" + f) for f in files
    ) as copies:
        assert_success("--check", *copies)
        gersemi("--in-place", *copies)
        assert_success("--check", *copies)


def test_format_in_place_multiple_not_formatted_files():
    files = ["file1.cmake", "file2.cmake", "file3.cmake"]
    with temporary_copies(
        case("/directory_with_not_formatted_files/" + f) for f in files
    ) as copies:
        assert_fail("--check", *copies)
        gersemi("--in-place", *copies)
        assert_success("--check", *copies)


def test_format_in_place_multiple_input_files_when_some_are_not_formatted():
    files = [
        "formatted_file1.cmake",
        "formatted_file2.cmake",
        "formatted_file3.cmake",
        "not_formatted_file1.cmake",
        "not_formatted_file2.cmake",
        "not_formatted_file3.cmake",
    ]
    with temporary_copies(
        case("/directory_with_some_not_formatted_files/" + f) for f in files
    ) as copies:
        formatted_copies, not_formatted_copies = copies[:3], copies[3:]
        assert_success("--check", *formatted_copies)
        assert_fail("--check", *not_formatted_copies)
        gersemi("--in-place", *formatted_copies, *not_formatted_copies)
        assert_success("--check", *formatted_copies, *not_formatted_copies)


def test_format_in_place_directory_with_formatted_files():
    with temporary_dir_copy(case("directory_with_formatted_files")) as copy:
        assert_success("--check", copy)
        gersemi("--in-place", copy)
        assert_success("--check", copy)


def test_format_in_place_directory_with_not_formatted_files():
    with temporary_dir_copy(case("directory_with_not_formatted_files")) as copy:
        assert_fail("--check", copy)
        gersemi("--in-place", copy)
        assert_success("--check", copy)


def test_format_in_place_directory_with_some_not_formatted_files():
    with temporary_dir_copy(case("directory_with_some_not_formatted_files")) as copy:
        assert_fail("--check", copy)
        gersemi("--in-place", copy)
        assert_success("--check", copy)


def test_format_with_default_line_length(tmpdir):
    inp = "set(FOO long_argument__________________________________________________________)"
    outp = inp + "\n"
    assert len(inp) == 80
    completed_process = gersemi("-", input=inp + "\n", cwd=tmpdir)
    assert completed_process.returncode == 0
    assert completed_process.stdout == outp
    assert completed_process.stderr == ""

    inp2 = "set(FOO long_argument____________________________________________________________)"
    outp2 = """set(FOO
    long_argument____________________________________________________________
)
"""
    assert len(inp2) > 80
    completed_process = gersemi("-", input=inp2 + "\n", cwd=tmpdir)
    assert completed_process.returncode == 0
    assert completed_process.stdout == outp2
    assert completed_process.stderr == ""


def test_format_with_non_default_line_length(tmpdir):
    line_length = 30
    inp = "set(FOO long_argument________)"
    outp = inp + "\n"
    assert len(inp) == line_length
    completed_process = gersemi(
        "--line-length",
        str(line_length),
        "-",
        input=inp + "\n",
        universal_newlines=True,
        cwd=tmpdir,
    )
    assert completed_process.returncode == 0
    assert completed_process.stdout == outp
    assert completed_process.stderr == ""

    inp2 = "set(FOO long_argument__________)"
    outp2 = """set(FOO
    long_argument__________
)
"""
    assert len(inp2) > line_length
    completed_process = gersemi(
        "--line-length",
        str(line_length),
        "-",
        input=inp2 + "\n",
        universal_newlines=True,
        cwd=tmpdir,
    )
    assert completed_process.returncode == 0
    assert completed_process.stdout == outp2
    assert completed_process.stderr == ""


def test_check_project_with_custom_commands():
    with temporary_dir_copy(case("custom_project/not_formatted")) as copy:
        assert_fail("--check", copy, "--definitions", copy)

    with temporary_dir_copy(case("custom_project/formatted")) as copy:
        assert_success("--check", copy, "--definitions", copy)


def test_check_project_with_custom_commands_but_without_definitions():
    with temporary_dir_copy(case("custom_project/not_formatted")) as copy:
        assert_fail("--check", copy)

    with temporary_dir_copy(
        case("custom_project/only_custom_commands_not_formatted")
    ) as copy:
        assert_success("--check", copy)


def test_format_project_with_custom_commands():
    with temporary_dir_copies(
        case("custom_project/" + d) for d in ["not_formatted", "formatted"]
    ) as (
        not_formatted,
        formatted,
    ):
        assert_that_directories_differ(not_formatted, formatted)
        assert_success("--check", formatted, "--definitions", formatted)
        assert_fail("--check", not_formatted, "--definitions", not_formatted)
        assert_success("--in-place", not_formatted, "--definitions", not_formatted)
        assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_format_project_with_custom_commands_but_without_definitions():
    with temporary_dir_copies(
        case("custom_project/" + d)
        for d in ["not_formatted", "only_custom_commands_not_formatted"]
    ) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)
        assert_success("--check", formatted)
        assert_fail("--check", not_formatted)
        assert_success("--in-place", not_formatted)
        assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_non_empty_stderr_when_files_are_not_formatted():
    completed_process = gersemi("--check", case("custom_project/not_formatted"))
    assert completed_process.returncode == 1
    assert completed_process.stderr != ""


def test_empty_stderr_when_files_are_not_formatted_but_quiet_is_supplied():
    completed_process = gersemi(
        "--check",
        case("custom_project/not_formatted"),
        "--quiet",
        universal_newlines=True,
    )
    assert completed_process.returncode == 1
    assert completed_process.stderr == ""


def test_project_with_dot_gersemirc_will_use_configuration_defined_in_file():
    with temporary_dir_copies(
        case("custom_project/" + d)
        for d in ["not_formatted", "formatted_with_line_length_100"]
    ) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)

        creator = lambda dirname: {"line_length": 100, "definitions": [dirname]}
        with create_configuration_files([not_formatted, formatted], creator):
            assert_success("--check", formatted)
            assert_fail("--check", not_formatted)
            assert_success("--in-place", not_formatted)

        assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_line_length_from_command_line_takes_precedence_over_configuration_file():
    with temporary_dir_copy(
        case("custom_project/formatted_with_line_length_100")
    ) as formatted:
        # without configuration file
        assert_fail("--check", formatted, "--definitions", formatted)
        assert_success(
            "--line-length", "100", "--check", formatted, "--definitions", formatted
        )
        assert_fail(
            "--line-length", "80", "--check", formatted, "--definitions", formatted
        )

        # with configuration file
        with create_dot_gersemirc(where=formatted, line_length=100):
            assert_success("--check", formatted, "--definitions", formatted)
            assert_success(
                "--line-length", "100", "--check", formatted, "--definitions", formatted
            )
            assert_fail(
                "--line-length", "80", "--check", formatted, "--definitions", formatted
            )

        # and without configuration file again
        assert_fail("--check", formatted, "--definitions", formatted)

        assert_success(
            "--line-length", "100", "--check", formatted, "--definitions", formatted
        )
        assert_fail(
            "--line-length", "80", "--check", formatted, "--definitions", formatted
        )


def test_definitions_from_command_line_take_precedence_over_configuration_file():
    directories = [case("custom_project/" + d) for d in ["not_formatted", "formatted"]]

    with temporary_dir_copies(directories) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)
        assert_success("--check", formatted, "--definitions", formatted)
        assert_fail("--check", not_formatted, "--definitions", not_formatted)
        assert_success("--in-place", not_formatted, "--definitions", not_formatted)
        assert_that_directories_have_the_same_content(not_formatted, formatted)

    # missing definitions in configuration file
    with temporary_dir_copies(directories) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)

        creator = lambda _: {}
        with create_configuration_files([not_formatted, formatted], creator):
            assert_success("--check", formatted)
            assert_fail("--check", not_formatted)
            assert_success("--in-place", not_formatted)

        assert_that_directories_differ(not_formatted, formatted)

    # missing definitions in configuration file but provided from command line
    with temporary_dir_copies(directories) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)

        creator = lambda _: {}
        with create_configuration_files([not_formatted, formatted], creator):
            assert_success("--check", formatted, "--definitions", formatted)
            assert_fail("--check", not_formatted, "--definitions", not_formatted)
            assert_success("--in-place", not_formatted, "--definitions", not_formatted)

        assert_that_directories_have_the_same_content(not_formatted, formatted)

    # best definitions are in configuration file but they are overriden by command line
    with temporary_dir_copies(directories) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)

        creator = lambda dirname: {"definitions": [dirname]}
        with create_configuration_files([not_formatted, formatted], creator):
            with ExitStack() as fake_definitions:
                fake_definitions_in_not_formatted, fake_definitions_in_formatted = [
                    fake_definitions.enter_context(
                        create_fake_definitions(where=dirname, name="definitions.cmake")
                    )
                    for dirname in [not_formatted, formatted]
                ]

                assert_success(
                    "--check", formatted, "--definitions", fake_definitions_in_formatted
                )
                assert_fail(
                    "--check",
                    not_formatted,
                    "--definitions",
                    fake_definitions_in_not_formatted,
                )
                assert_success(
                    "--in-place",
                    not_formatted,
                    "--definitions",
                    fake_definitions_in_not_formatted,
                )

        assert_that_directories_differ(not_formatted, formatted)


def test_use_paths_relative_to_root_as_definitions_in_configuration_file():
    with temporary_dir_copies(
        case("custom_project/" + d)
        for d in ["not_formatted", "formatted_with_line_length_100"]
    ) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)

        creator = lambda dirname: {
            "line_length": 100,
            "definitions": [
                "back_to_the_future.cmake",
                "./back_to_the_future_sequels.cmake",
            ],
        }
        with create_configuration_files([not_formatted, formatted], creator):
            assert_success("--check", formatted)
            assert_fail("--check", not_formatted)
            assert_success("--in-place", not_formatted)

        assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_use_absolute_paths_as_definitions_in_configuration_file():
    with temporary_dir_copies(
        case("custom_project/" + d)
        for d in ["not_formatted", "formatted_with_line_length_100"]
    ) as (not_formatted, formatted):
        assert_that_directories_differ(not_formatted, formatted)

        creator = lambda dirname: {
            "line_length": 100,
            "definitions": [
                os.path.join(dirname, "back_to_the_future.cmake"),
                os.path.join(dirname, "back_to_the_future_sequels.cmake"),
            ],
        }
        with create_configuration_files([not_formatted, formatted], creator):
            assert_success("--check", formatted)
            assert_fail("--check", not_formatted)
            assert_success("--in-place", not_formatted)

        assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_use_configuration_file_from_current_directory_when_input_is_from_stdin(tmpdir):
    line_length = 30
    inp = "set(FOO long_argument__________)"
    outp = """set(FOO
    long_argument__________
)
"""
    assert len(inp) > line_length

    with create_dot_gersemirc(where=tmpdir, line_length=30):
        completed_process = gersemi("-", input=inp + "\n", cwd=tmpdir)
        assert completed_process.returncode == 0
        assert completed_process.stdout == outp
        assert completed_process.stderr == ""


def test_use_configuration_file_from_parent_directory_when_input_is_from_stdin(tmpdir):
    line_length = 30
    inp = "set(FOO long_argument__________)"
    outp = """set(FOO
    long_argument__________
)
"""
    assert len(inp) > line_length

    with create_dot_gersemirc(where=tmpdir, line_length=30):
        nested_directory = tmpdir.mkdir("foo").mkdir("bar")
        completed_process = gersemi("-", input=inp + "\n", cwd=nested_directory)
        assert completed_process.returncode == 0
        assert completed_process.stdout == outp
        assert completed_process.stderr == ""


@contextmanager
def cache_tests(files_to_format):
    with ExitStack() as stack:
        copy, temporary_cache = [
            stack.enter_context(temporary_dir_copy(files_to_format)),
            stack.enter_context(make_temporary_cache()),
        ]
        gersemi_ = lambda *args, **kwargs: gersemi_with_cache_path(
            temporary_cache, *args, **kwargs
        )
        inspector = stack.enter_context(inspect_cache(temporary_cache))
        yield copy, gersemi_, inspector


def test_formatted_files_are_stored_in_cache_on_check():
    with cache_tests(case("custom_project/formatted")) as (copy, gersemi_, inspector):
        inspector.assert_that_has_no_tables()

        gersemi_("--check", copy, "--definitions", copy)

        inspector.assert_that_has_initialized_tables()

        assert len(inspector.get_files()) > 0
        assert len(inspector.get_formatted()) > 0


def test_not_formatted_files_are_not_stored_in_cache_on_check():
    with cache_tests(case("custom_project/not_formatted")) as (
        copy,
        gersemi_,
        inspector,
    ):
        inspector.assert_that_has_no_tables()

        gersemi_("--check", copy, "--definitions", copy)

        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) == 0
        assert len(inspector.get_formatted()) == 0


def test_not_formatted_files_are_stored_in_cache_after_formatting():
    with cache_tests(case("custom_project/not_formatted")) as (
        copy,
        gersemi_,
        inspector,
    ):
        inspector.assert_that_has_no_tables()

        gersemi_("--in-place", copy, "--definitions", copy)

        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) > 0
        assert len(inspector.get_formatted()) > 0


def test_formatted_files_in_cache_dont_get_updated_on_subsequent_run():
    with cache_tests(case("custom_project/not_formatted")) as (
        copy,
        gersemi_,
        inspector,
    ):
        inspector.assert_that_has_no_tables()

        gersemi_("--in-place", copy, "--definitions", copy)
        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) > 0
        formatted_after_first_run = inspector.get_formatted()
        assert len(formatted_after_first_run) > 0

        gersemi_("--in-place", copy, "--definitions", copy)
        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) > 0
        formatted_after_second_run = inspector.get_formatted()
        assert len(formatted_after_second_run) > 0

        assert formatted_after_first_run == formatted_after_second_run


def test_different_configuration_leads_to_overriding_data_stored_in_cache():
    with cache_tests(case("custom_project/not_formatted")) as (
        copy,
        gersemi_,
        inspector,
    ):
        inspector.assert_that_has_no_tables()

        gersemi_("--in-place", copy, "--definitions", copy)
        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) > 0
        formatted_after_first_run = inspector.get_formatted()
        assert len(formatted_after_first_run) > 0

        gersemi_("--in-place", copy, "--definitions", copy, "--line-length", "100")
        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) > 0
        formatted_after_second_run = inspector.get_formatted()
        assert len(formatted_after_first_run) > 0

        only_paths_from_first_run = [path for (path, *_) in formatted_after_first_run]
        only_paths_from_second_run = [path for (path, *_) in formatted_after_second_run]

        assert only_paths_from_first_run == only_paths_from_second_run
        assert formatted_after_first_run != formatted_after_second_run


def test_no_files_are_stored_in_cache_on_diff():
    with cache_tests(case("custom_project/not_formatted")) as (
        copy,
        gersemi_,
        inspector,
    ):
        inspector.assert_that_has_no_tables()
        gersemi_("--diff", copy, "--definitions", copy)
        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) == 0
        assert len(inspector.get_formatted()) == 0


def test_when_cache_cant_be_modified_it_is_ignored():
    with temporary_dir_copy(
        case("custom_project/formatted")
    ) as copy, tempfile.NamedTemporaryFile(mode="r") as cache_path:
        os.chmod(cache_path.name, S_IREAD | S_IRGRP | S_IROTH)

        gersemi_ = lambda *args, **kwargs: gersemi_with_cache_path(
            cache_path.name, *args, **kwargs
        )

        completed_process = gersemi_("--check", copy, "--definitions", copy)
        assert completed_process.returncode == 0
        assert completed_process.stdout == ""
        assert completed_process.stderr == ""


def test_when_cache_is_malformed_it_is_ignored():
    with temporary_dir_copy(
        case("custom_project/formatted")
    ) as copy, tempfile.NamedTemporaryFile() as cache_path:
        cache_path.write("foobarbaz1231212312312312313".encode("ascii"))
        cache_path.flush()

        gersemi_ = lambda *args, **kwargs: gersemi_with_cache_path(
            cache_path.name, *args, **kwargs
        )

        completed_process = gersemi_("--check", copy, "--definitions", copy)
        assert completed_process.returncode == 0
        assert completed_process.stdout == ""
        assert completed_process.stderr == ""


def test_cache_is_not_updated_when_input_is_from_stdin():
    inp = """set(FOO BAR)
"""
    with make_temporary_cache() as cache_path, inspect_cache(cache_path) as inspector:
        gersemi_ = lambda *args, **kwargs: gersemi_with_cache_path(
            cache_path, *args, **kwargs
        )

        inspector.assert_that_has_no_tables()
        gersemi_("--check", "-", input=inp)

        # first use initialized tables in cache
        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) == 0
        assert len(inspector.get_formatted()) == 0

        gersemi_("--check", "-", input=inp)
        inspector.assert_that_has_initialized_tables()
        assert len(inspector.get_files()) == 0
        assert len(inspector.get_formatted()) == 0


def test_check_project_with_conflicting_command_definitions():
    with temporary_dir_copy(case("conflicting_definitions")) as copy:
        completed_process = gersemi("--check", copy, "--definitions", copy)
        assert completed_process.returncode == 0
        assert completed_process.stdout == ""
        assert re.search(
            f"""Warning: conflicting definitions for 'foo':
\\(used\\)    .*foo1\\.cmake:1:10
\\(ignored\\) .*foo2\\.cmake:1:10
\\(ignored\\) .*foo2\\.cmake:5:10
\\(ignored\\) .*foo2\\.cmake:9:10
""",
            completed_process.stderr,
            re.MULTILINE,
        )


def test_format_file_with_conflicting_command_definitions():
    with temporary_dir_copy(case("conflicting_definitions")) as copy:
        completed_process = gersemi(
            f"{copy}/CMakeLists.txt",
            "--definitions",
            copy,
            "--list-expansion=favour-expansion",
        )
        assert completed_process.returncode == 0
        assert (
            completed_process.stdout
            == """foo(ONE
    TWO x
    THREE
        y
        z
)
"""
        )
        assert re.search(
            f"""Warning: conflicting definitions for 'foo':
\\(used\\)    .*foo1\\.cmake:1:10
\\(ignored\\) .*foo2\\.cmake:1:10
\\(ignored\\) .*foo2\\.cmake:5:10
\\(ignored\\) .*foo2\\.cmake:9:10
""",
            completed_process.stderr,
            re.MULTILINE,
        )



from typing import List
from lark import Tree
from lark.visitors import Transformer, TransformerChain, Transformer_InPlace
from gersemi.ast_helpers import contains_line_comment, is_one_of_keywords
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.types import Nodes
from gersemi.utils import advance


class IsolateUnaryOperators(Transformer_InPlace):
    unary_operators: List[str] = []

    def arguments(self, children: Nodes) -> Tree:
        if len(children) < 2:
            return Tree("arguments", children)

        new_children: Nodes = []
        iterator = zip(children, children[1:])
        is_one_of_unary_operators = is_one_of_keywords(self.unary_operators)
        for one_behind, current in iterator:
            if is_one_of_unary_operators(one_behind):
                new_children += [Tree("unary_operation", [one_behind, current])]
                _, current = advance(iterator, times=1, default=(None, None))
                if current is None:
                    break
            else:
                new_children += [one_behind]
        else:
            new_children += [item for item in [current] if item is not None]
        return Tree("arguments", new_children)


class IsolateUnaryTests(IsolateUnaryOperators):
    unary_operators: List[str] = [
        "COMMAND",
        "POLICY",
        "TARGET",
        "TEST",
        "EXISTS",
        "IS_DIRECTORY",
        "IS_SYMLINK",
        "IS_ABSOLUTE",
        "DEFINED",
    ]


class IsolateBinaryTests(Transformer_InPlace):
    binary_operators: List[str] = [
        "IS_NEWER_THAN",
        "MATCHES",
        "LESS",
        "GREATER",
        "EQUAL",
        "LESS_EQUAL",
        "GREATER_EQUAL",
        "STRLESS",
        "STRGREATER",
        "STREQUAL",
        "STRLESS_EQUAL",
        "STRGREATER_EQUAL",
        "VERSION_LESS",
        "VERSION_GREATER",
        "VERSION_EQUAL",
        "VERSION_LESS_EQUAL",
        "VERSION_GREATER_EQUAL",
        "IN_LIST",
        "PATH_EQUAL",
    ]

    def arguments(self, children: Nodes) -> Tree:
        if len(children) < 3:
            return Tree("arguments", children)

        new_children: Nodes = []
        iterator = zip(children, children[1:], children[2:])
        is_one_of_binary_operators = is_one_of_keywords(self.binary_operators)
        for two_behind, one_behind, current in iterator:
            if is_one_of_binary_operators(one_behind):
                new_children += [
                    Tree("binary_operation", [two_behind, one_behind, current])
                ]
                _, one_behind, current = advance(
                    iterator, times=2, default=(None, None, None)
                )
                if current is None:
                    break
            else:
                new_children += [two_behind]
        else:
            new_children += [item for item in [one_behind, current] if item is not None]
        return Tree("arguments", new_children)


class IsolateNotExpressions(IsolateUnaryOperators):
    unary_operators: List[str] = ["NOT"]


class IsolateAndExpressions(IsolateUnaryOperators):
    unary_operators: List[str] = ["AND"]


class IsolateOrExpressions(IsolateUnaryOperators):
    unary_operators: List[str] = ["OR"]


def IsolateConditions() -> Transformer:
    return TransformerChain(
        IsolateUnaryTests(),
        IsolateBinaryTests(),
        IsolateNotExpressions(),
        IsolateAndExpressions(),
        IsolateOrExpressions(),
    )


class ConditionSyntaxCommandInvocationDumper(BaseCommandInvocationDumper):
    inhibit_favour_expansion = True

    def unary_operation(self, tree):
        result = self._try_to_format_into_single_line(tree.children, separator=" ")
        if result is not None:
            return result

        operation, arg = tree.children
        formatted_operation = self.visit(operation)

        if (not contains_line_comment([operation])) and (
            len(formatted_operation.strip()) < self.indent_size
        ):
            return f"{formatted_operation} {self.visit(arg).lstrip()}"

        with self.indented():
            formatted_arg = self.visit(arg)
        return f"{formatted_operation}\n{formatted_arg}"

    def binary_operation(self, tree):
        result = self._try_to_format_into_single_line(tree.children, separator=" ")
        if result is not None:
            return result

        lhs, operation, rhs = tree.children
        formatted_lhs = self.visit(lhs)
        with self.indented():
            formatted_operation = self.visit(operation)
            formatted_rhs = self.visit(rhs)
        return f"{formatted_lhs}\n{formatted_operation}\n{formatted_rhs}"

    def arguments(self, tree):
        preprocessed = IsolateConditions().transform(tree)
        return super().arguments(preprocessed)

    def _preprocess_arguments(self, arguments):
        return IsolateConditions().transform(arguments)

    def _split_arguments(self, arguments):
        if len(arguments) < 1:
            return arguments

        head, *tail = arguments
        return [[head], tail]

    def complex_argument(self, tree):
        arguments, *_ = tree.children
        result = self._try_to_format_into_single_line(
            arguments.children, separator=" ", prefix="(", postfix=")"
        )
        if result is not None:
            return result

        begin = self._indent("(\n")
        with self.indented():
            formatted_arguments = self.visit(arguments)
        end = self._indent(")")
        return f"{begin}{formatted_arguments}\n{end}"



from typing import Dict, Iterator, Iterable, List, Sequence, Sized, Tuple
from gersemi.ast_helpers import is_one_of_keywords, is_comment
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.types import Nodes
from gersemi.utils import pop_all


def to_list_of_single_item_lists(sequence):
    return [*map(lambda item: [item], sequence)]


def is_non_empty(sequence: Sized) -> bool:
    return len(sequence) > 0


def is_empty(sequence: Sized) -> bool:
    return len(sequence) <= 0


class PositionalArguments(list):
    pass


class KeywordSplitter:
    def __init__(self, options, one_value_keywords, multi_value_keywords):
        self.is_one_of_options = is_one_of_keywords(options)
        self.is_one_of_one_value_keywords = is_one_of_keywords(one_value_keywords)
        self.is_one_of_multi_value_keywords = is_one_of_keywords(multi_value_keywords)
        self.groups: List[Nodes] = []
        self.accumulator: Nodes = []
        self.comment_accumulator: Nodes = []

    def _flush_accumulators(self):
        each_comment_in_its_own_group = [[c] for c in pop_all(self.comment_accumulator)]
        if isinstance(self.accumulator, PositionalArguments):
            argument_group = PositionalArguments(pop_all(self.accumulator))
        else:
            argument_group = pop_all(self.accumulator)
        self.groups += [argument_group, *each_comment_in_its_own_group]

    def _append_option_group(self, argument):
        self._flush_accumulators()
        self.groups += [[argument]]

    def _append_one_value_group(self, argument, iterator):
        self._flush_accumulators()
        group = [argument]
        for n in iterator:
            group.append(n)
            if not is_comment(n):
                break
        self.groups.append(group)

    def split(self, arguments: Nodes) -> Iterator[Nodes]:
        iterator = iter(arguments)
        for argument in iterator:
            if is_comment(argument):
                self.comment_accumulator += [argument]
            elif self.is_one_of_options(argument):
                self._append_option_group(argument)
            elif self.is_one_of_one_value_keywords(argument):
                self._append_one_value_group(argument, iterator)
            elif self.is_one_of_multi_value_keywords(argument):
                self._flush_accumulators()
                self.accumulator = [argument]
            elif is_non_empty(self.accumulator):
                self.accumulator += [*pop_all(self.comment_accumulator), argument]
            else:
                if not isinstance(self.accumulator, PositionalArguments):
                    self.accumulator = PositionalArguments()
                self.accumulator += [*pop_all(self.comment_accumulator), argument]

        self._flush_accumulators()
        return filter(is_non_empty, self.groups)


class ArgumentAwareCommandInvocationDumper(BaseCommandInvocationDumper):
    inhibit_favour_expansion: bool = False
    front_positional_arguments: Sequence[str] = []
    back_positional_arguments: Sequence[str] = []
    options: Iterable[str] = []
    one_value_keywords: Iterable[str] = []
    multi_value_keywords: Iterable[str] = []
    keyword_formatters: Dict[str, str] = {}

    def _default_format_values(self, values) -> str:
        return "\n".join(map(self.visit, values))

    def _format_positional_arguments_group(self, group) -> str:
        return "\n".join(map(self.visit, group))

    def _format_group(self, group) -> str:
        if isinstance(group, PositionalArguments):
            return self._format_positional_arguments_group(group)

        result = self._try_to_format_into_single_line(group, separator=" ")
        if result is not None:
            return result

        keyword, *values = group
        keyword_as_value = keyword.children[0] if len(keyword.children) > 0 else None

        can_be_inlined = (not self.favour_expansion) or (
            self.favour_expansion
            and keyword is not None
            and keyword_as_value not in self.multi_value_keywords
        )
        if can_be_inlined:
            with self.select_inlining_strategy():
                result = self._try_to_format_into_single_line(group, separator=" ")
                if result is not None:
                    return result

        begin = self.visit(keyword)
        if len(values) == 0:
            return begin

        with self.indented():
            formatter = getattr(
                self,
                self.keyword_formatters.get(keyword_as_value, "_default_format_values"),
            )
            formatted_values = formatter(values)
        return f"{begin}\n{formatted_values}"

    def _split_positional_arguments(self, arguments: Nodes, positional_arguments):
        last_index = min(len(arguments), len(positional_arguments))
        result = []
        for i in range(last_index):
            result.append(PositionalArguments([arguments[i]]))

        rest = arguments[last_index:]
        if len(rest) > 0:
            result.append(PositionalArguments(rest))
        return result

    def _separate_front(self, arguments: Nodes) -> Tuple[List[Nodes], Nodes]:
        is_keyword = is_one_of_keywords(
            list(self.options)
            + list(self.one_value_keywords)
            + list(self.multi_value_keywords)
        )
        for index, argument in enumerate(arguments):
            if is_keyword(argument):
                pivot = index
                break
        else:
            return (
                self._split_positional_arguments(
                    arguments, self.front_positional_arguments
                ),
                [],
            )
        return (
            self._split_positional_arguments(
                arguments[:pivot], self.front_positional_arguments
            ),
            arguments[pivot:],
        )

    def _split_by_keywords(self, arguments: Nodes) -> Iterator[Nodes]:
        splitter = KeywordSplitter(
            self.options,
            self.one_value_keywords,
            self.multi_value_keywords,
        )
        return splitter.split(arguments)

    def _split_arguments(self, arguments):
        if len(self.back_positional_arguments) > 0:
            arguments, back = (
                arguments[: -len(self.back_positional_arguments)],
                [
                    PositionalArguments(
                        arguments[-len(self.back_positional_arguments) :]
                    )
                ],
            )
        else:
            back = []
        front, tail = self._separate_front(arguments)
        keyworded_arguments = self._split_by_keywords(tail)

        return [*front, *keyworded_arguments, *back]

    def group_size(self, group):
        if isinstance(group, PositionalArguments):
            return len(group)
        return max(0, len(group) - 1)

    def arguments(self, tree):
        groups = self._split_arguments(tree.children)
        return "\n".join(map(self._format_group, filter(None, groups)))



from gersemi.configuration import ListExpansion
from gersemi.dumper import Dumper
from gersemi.parser import BARE_PARSER, PARSER
from gersemi.sanity_checker import check_code_equivalence


def noop(*_):
    pass


class Formatter:
    def __init__(
        self,
        sanity_checker,
        line_length: int,
        custom_command_definitions,
        list_expansion: ListExpansion,
    ):
        self.sanity_checker = sanity_checker
        self.line_length = line_length
        self.custom_command_definitions = custom_command_definitions
        self.list_expansion = list_expansion

    def format(self, code):
        dumper = Dumper(
            self.line_length, self.custom_command_definitions, self.list_expansion
        )
        result = dumper.visit(PARSER.parse(code))
        self.sanity_checker(BARE_PARSER, code, result)
        return result


def create_formatter(
    do_sanity_check, line_length, custom_command_definitions, list_expansion
):
    sanity_checker = check_code_equivalence if do_sanity_check else noop
    return Formatter(
        sanity_checker, line_length, custom_command_definitions, list_expansion
    )



from dataclasses import dataclass
from typing import Tuple


@dataclass
class Keywords:
    options: Tuple[str, ...] = tuple()
    one_value_keywords: Tuple[str, ...] = tuple()
    multi_value_keywords: Tuple[str, ...] = tuple()



from typing import Sequence, Tuple
from lark import Tree
from lark.visitors import Transformer, Transformer_InPlace
from gersemi.ast_helpers import is_keyword
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.utils import advance


class PrependLhs(Transformer):
    def __init__(self, lhs):
        super().__init__()
        self.lhs = lhs

    def unquoted_argument(self, children):
        return Tree("unquoted_argument", [self.lhs + " " + "".join(children)])


class IsolateTwoWordKeywords(Transformer_InPlace):
    def __init__(self, lhs, rhs):
        super().__init__()
        self.lhs = lhs
        self.rhs = rhs

    def _is_lhs(self, node):
        return is_keyword(self.lhs)(node)

    def _is_rhs(self, node):
        return is_keyword(self.rhs)(node)

    def _prepend_lhs(self, node):
        return PrependLhs(self.lhs).transform(node)

    def arguments(self, children):
        if len(children) <= 1:
            return Tree("arguments", children)

        new_children = []
        iterator = zip(children, children[1:])
        for one_behind, current in iterator:
            if self._is_lhs(one_behind) and self._is_rhs(current):
                new_children += [self._prepend_lhs(current)]
                _, current = advance(iterator, times=1, default=(None, None))
                if current is None:
                    break
            else:
                new_children += [one_behind]
        else:
            new_children += [item for item in [current] if item is not None]
        return Tree("arguments", new_children)


class TwoWordKeywordIsolator(BaseCommandInvocationDumper):
    two_words_keywords: Sequence[Tuple[str, str]] = []

    def _preprocess_arguments(self, arguments):
        preprocessed = arguments
        for lhs, rhs in self.two_words_keywords:
            preprocessed = IsolateTwoWordKeywords(lhs, rhs).transform(preprocessed)
        return preprocessed



from contextlib import contextmanager
from itertools import filterfalse
from typing import Dict, List, Optional, Union
from lark import Tree
from gersemi.ast_helpers import is_comment, is_unquoted_argument
from gersemi.types import Nodes
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class MultipleSignatureCommandInvocationDumper(ArgumentAwareCommandInvocationDumper):
    customized_signatures: Dict[Optional[str], Dict[str, Union[List, int]]] = {}

    @contextmanager
    def _update_signature_characteristics(self, signature):
        if signature is None:
            yield
            return

        try:

            def get(key):
                return signature.get(key, [])

            self.front_positional_arguments = get("front_positional_arguments")
            self.back_positional_arguments = get("back_positional_arguments")
            self.options = get("options")
            self.one_value_keywords = get("one_value_keywords")
            self.multi_value_keywords = get("multi_value_keywords")
            yield
        finally:
            delattr(self, "front_positional_arguments")
            delattr(self, "back_positional_arguments")
            delattr(self, "options")
            delattr(self, "one_value_keywords")
            delattr(self, "multi_value_keywords")

    def format_command(self, tree):
        _, arguments = tree.children
        arguments = self._preprocess_arguments(arguments)
        arguments_only = filterfalse(is_comment, arguments.children)
        first_argument = next(arguments_only, None)
        if first_argument is None or not is_unquoted_argument(first_argument):
            signature = self.customized_signatures.get(None, None)
            if signature is None:
                return super().format_command(tree)
            with self._update_signature_characteristics(signature):
                return super().format_command(tree)

        first_argument_as_value = first_argument.children[0]
        signature = self.customized_signatures.get(first_argument_as_value, None)
        if signature is None:
            signature = self.customized_signatures.get(None, None)

        with self._update_signature_characteristics(signature):
            return super().format_command(tree)

    def _split_arguments(self, arguments: Nodes) -> List[Nodes]:
        signature_node, *rest = arguments
        if isinstance(signature_node, Tree):
            if len(signature_node.children) > 0:
                signature_node_as_value = str(signature_node.children[0])
                if (
                    (signature_node_as_value in self.options)
                    or (signature_node_as_value in self.one_value_keywords)
                    or (signature_node_as_value in self.multi_value_keywords)
                ):
                    return super()._split_arguments(arguments)
        return [[signature_node], *super()._split_arguments(rest)]



from contextlib import contextmanager
from typing import Dict, Iterable
from gersemi.ast_helpers import is_commented_argument
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class SectionAwareCommandInvocationDumper(ArgumentAwareCommandInvocationDumper):
    sections: Dict[str, Dict[str, Iterable[str]]] = {}

    @contextmanager
    def _update_section_characteristics(self, keyword):
        try:
            section = self.sections[keyword]
            self.options = section.get("options", [])
            self.one_value_keywords = section.get("one_value_keywords", [])
            self.multi_value_keywords = section.get("multi_value_keywords", [])
            yield
        finally:
            delattr(self, "options")
            delattr(self, "one_value_keywords")
            delattr(self, "multi_value_keywords")

    def _format_group(self, group) -> str:
        result = self._try_to_format_into_single_line(group, separator=" ")
        if result is not None:
            return result

        original_format_group = super()._format_group
        front_node, *rest = group
        if is_commented_argument(front_node):
            inner, *_ = front_node.children
            keyword, *_ = inner.children
        else:
            keyword, *_ = front_node.children

        if keyword not in self.sections:
            return original_format_group(group)

        with self._update_section_characteristics(keyword):
            subgroups = self._split_arguments(rest)
            formatted_front = original_format_group([front_node])
            with self.indented():
                return "\n".join(
                    [formatted_front, *map(original_format_group, subgroups)]
                )



from typing import List
from gersemi.ast_helpers import contains_line_comment
from gersemi.base_dumper import BaseDumper
from gersemi.configuration import ListExpansion
from gersemi.types import Nodes


class BaseCommandInvocationDumper(BaseDumper):
    def format_command_with_short_name(self, begin, arguments, end):
        with self.indented():
            formatted_arguments = self.visit(arguments).lstrip()
        if (
            not contains_line_comment(arguments.children)
            and "\n" not in formatted_arguments
        ):
            return "".join([self._indent(begin), formatted_arguments, end])

        return f"{self._indent(begin)}{formatted_arguments}\n{self._indent(end)}"

    def _format_command_with_long_name(self, begin, arguments, end):
        with self.indented():
            formatted_arguments = self.visit(arguments)
        return "\n".join([self._indent(begin), formatted_arguments, self._indent(end)])

    def _split_arguments(self, arguments: Nodes) -> List[Nodes]:
        return [arguments]

    def group_size(self, group):
        return len(group)

    def _inlining_condition(self, arguments):
        groups = self._split_arguments(arguments.children)
        group_sizes = list(map(self.group_size, groups))
        if (
            self.list_expansion == ListExpansion.FavourExpansion
            and not self.inhibit_favour_expansion
        ):
            return all(size < 2 for size in group_sizes)
        return all(size <= 4 for size in group_sizes)

    def format_command(self, tree):
        identifier, arguments = tree.children
        arguments = self._preprocess_arguments(arguments)
        begin = f"{identifier}("
        end = ")"
        if self._inlining_condition(arguments):
            result = self._try_to_format_into_single_line(
                arguments.children, separator=" ", prefix=f"{identifier}(", postfix=")"
            )
            if result is not None:
                return result

        with self.select_expansion_strategy():
            if len(begin) == self.indent_size:
                return self.format_command_with_short_name(begin, arguments, end)
            return self._format_command_with_long_name(begin, arguments, end)

    def arguments(self, tree):
        return "\n".join(self.visit_children(tree))

    def commented_argument(self, tree):
        argument, comment, *_ = tree.children
        formatted_argument = self.visit(argument)
        with self.not_indented():
            formatted_comment = self.visit(comment)
        return f"{formatted_argument} {formatted_comment}"

    def complex_argument(self, tree):
        arguments, *_ = tree.children
        if len(arguments.children) <= 4:
            result = self._try_to_format_into_single_line(
                arguments.children, separator=" ", prefix="(", postfix=")"
            )
            if result is not None:
                return result

        begin = self._indent("(\n")
        with self.indented():
            formatted_arguments = self.visit(arguments)
        end = self._indent(")")
        return f"{begin}{formatted_arguments}\n{end}"

    def bracket_comment(self, tree):
        return " " * self.alignment + "#" + self.__default__(tree)

    def bracket_argument(self, tree):
        return " " * self.alignment + self.__default__(tree)

    def quoted_argument(self, tree):
        return " " * self.alignment + f'"{self.__default__(tree)}"'

    def unquoted_argument(self, tree):
        return self._indent(self.__default__(tree))

    def _preprocess_arguments(self, arguments):
        return arguments



from gersemi.command_line_formatter import CommandLineFormatter
from gersemi.keyword_with_pairs_formatter import KeywordWithPairsFormatter
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)
from .multiple_signature_command_invocation_dumper import (
    MultipleSignatureCommandInvocationDumper,
)
from .section_aware_command_invocation_dumper import SectionAwareCommandInvocationDumper
from .target_link_libraries_command_dumper import TargetLinkLibraries
from .two_word_keyword_isolator import TwoWordKeywordIsolator


class AddCustomCommand(CommandLineFormatter, MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        "OUTPUT": dict(
            options=[
                "VERBATIM",
                "APPEND",
                "USES_TERMINAL",
                "COMMAND_EXPAND_LISTS",
                "DEPENDS_EXPLICIT_ONLY",
            ],
            one_value_keywords=[
                "MAIN_DEPENDENCY",
                "WORKING_DIRECTORY",
                "COMMENT",
                "DEPFILE",
                "JOB_POOL",
            ],
            multi_value_keywords=[
                "OUTPUT",
                "COMMAND",
                "ARGS",
                "DEPENDS",
                "BYPRODUCTS",
                "IMPLICIT_DEPENDS",
                "OUTPUT",
            ],
        ),
        "TARGET": dict(
            options=[
                "PRE_BUILD",
                "PRE_LINK",
                "POST_BUILD",
                "VERBATIM",
                "USES_TERMINAL",
                "COMMAND_EXPAND_LISTS",
            ],
            one_value_keywords=["TARGET", "WORKING_DIRECTORY", "COMMENT", "TARGET"],
            multi_value_keywords=["COMMAND", "ARGS", "BYPRODUCTS"],
        ),
    }
    keyword_formatters = {
        "COMMAND": "_format_command_line",
        "ARGS": "_format_command_line",
    }


class AddCustomTarget(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["Name", "ALL"]
    options = ["VERBATIM", "USES_TERMINAL", "COMMAND_EXPAND_LISTS"]
    one_value_keywords = ["WORKING_DIRECTORY", "COMMENT", "JOB_POOL"]
    multi_value_keywords = ["COMMAND", "DEPENDS", "BYPRODUCTS", "SOURCES"]
    keyword_formatters = {"COMMAND": "_format_command_line"}

    def _format_positional_arguments_group(self, group):
        if len(group) > 1:
            if group[0].children[0] == "ALL":
                first, *rest = group
                return f"{self.visit(first)}\n{super()._format_command_line(rest)}"
        return super()._format_command_line(group)


class AddDependencies(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]


class AddExecutable(ArgumentAwareCommandInvocationDumper):
    two_words_keywords = [("IMPORTED", "GLOBAL")]
    front_positional_arguments = ["<name>"]
    options = [
        "WIN32",
        "MACOSX_BUNDLE",
        "EXCLUDE_FROM_ALL",
        "IMPORTED",
        "IMPORTED GLOBAL",
    ]
    one_value_keywords = ["ALIAS"]


class AddLibrary(TwoWordKeywordIsolator, ArgumentAwareCommandInvocationDumper):
    two_words_keywords = [("IMPORTED", "GLOBAL")]
    front_positional_arguments = ["<name>"]
    options = [
        "STATIC",
        "SHARED",
        "MODULE",
        "EXCLUDE_FROM_ALL",
        "OBJECT",
        "IMPORTED",
        "IMPORTED GLOBAL",
        "UNKNOWN",
    ]
    one_value_keywords = ["ALIAS"]
    multi_value_keywords = ["INTERFACE"]


class AddSubdirectory(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["source_dir", "binary_dir"]
    options = ["EXCLUDE_FROM_ALL", "SYSTEM"]


class AddTest(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<name>", "<command>"]
    options = ["COMMAND_EXPAND_LISTS"]
    one_value_keywords = ["NAME", "WORKING_DIRECTORY"]
    multi_value_keywords = ["COMMAND", "CONFIGURATIONS"]
    keyword_formatters = {"COMMAND": "_format_command_line"}


class BuildCommand(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>"]
    one_value_keywords = ["CONFIGURATION", "TARGET", "PROJECT_NAME", "PARALLEL_LEVEL"]


class CMakeFileApi(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        "QUERY": dict(
            options=["QUERY"],
            one_value_keywords=["API_VERSION"],
            multi_value_keywords=["CODEMODEL", "CACHE", "CMAKEFILES", "TOOLCHAINS"],
        ),
    }


class CreateTestSourcelist(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["sourceListName", "driverName"]
    one_value_keywords = ["EXTRA_INCLUDE", "FUNCTION"]


class DefineProperty(ArgumentAwareCommandInvocationDumper):
    options = [
        "GLOBAL",
        "DIRECTORY",
        "TARGET",
        "SOURCE",
        "TEST",
        "VARIABLE",
        "CACHED_VARIABLE",
        "INHERITED",
    ]
    one_value_keywords = ["PROPERTY", "INITIALIZE_FROM_VARIABLE"]
    multi_value_keywords = ["BRIEF_DOCS", "FULL_DOCS"]


class Export(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        "EXPORT": dict(one_value_keywords=["EXPORT", "NAMESPACE", "FILE"]),
        "TARGETS": dict(
            options=["APPEND", "EXPORT_LINK_INTERFACE_LIBRARIES"],
            one_value_keywords=["NAMESPACE", "FILE", "ANDROID_MK"],
            multi_value_keywords=["TARGETS"],
        ),
        "PACKAGE": dict(one_value_keywords=["PACKAGE"]),
    }


class FtlkWrapUi(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["resultingLibraryName"]


class GetSourceFileProperty(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>", "<file>"]
    back_positional_arguments = ["<property>"]
    one_value_keywords = ["DIRECTORY", "TARGET_DIRECTORY"]


class GetTargetProperty(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<VAR>", "target", "property"]


class GetTestProperty(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<VAR>", "target", "property"]


class IncludeDirectories(ArgumentAwareCommandInvocationDumper):
    options = ["AFTER", "BEFORE", "SYSTEM"]


class IncludeExternalMsProject(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["projectname", "location"]
    one_value_keywords = ["TYPE", "GUID", "PLATFORM"]


class Install(TwoWordKeywordIsolator, MultipleSignatureCommandInvocationDumper):
    two_words_keywords = [("INCLUDES", "DESTINATION")]

    customized_signatures = {
        "TARGETS": dict(
            options=[
                "ARCHIVE",
                "LIBRARY",
                "RUNTIME",
                "OBJECTS",
                "FRAMEWORK",
                "BUNDLE",
                "PRIVATE_HEADER",
                "PUBLIC_HEADER",
                "RESOURCE",
                "OPTIONAL",
                "EXCLUDE_FROM_ALL",
                "NAMELINK_ONLY",
                "NAMELINK_SKIP",
            ],
            one_value_keywords=[
                "EXPORT",
                "DESTINATION",
                "COMPONENT",
                "NAMELINK_COMPONENT",
                "RUNTIME_DEPENDENCY_SET",
                "FILE_SET",
            ],
            multi_value_keywords=[
                "TARGETS",
                "PERMISSIONS",
                "CONFIGURATIONS",
                "INCLUDES DESTINATION",
                "RUNTIME_DEPENDENCIES",
            ],
        ),
        "FILES": dict(
            options=["OPTIONAL", "EXCLUDE_FROM_ALL"],
            one_value_keywords=["TYPE", "DESTINATION", "COMPONENT", "RENAME"],
            multi_value_keywords=["FILES", "PERMISSIONS", "CONFIGURATIONS"],
        ),
        "PROGRAMS": dict(
            options=["OPTIONAL", "EXCLUDE_FROM_ALL"],
            one_value_keywords=["TYPE", "DESTINATION", "COMPONENT", "RENAME"],
            multi_value_keywords=["PROGRAMS", "PERMISSIONS", "CONFIGURATIONS"],
        ),
        "DIRECTORY": dict(
            options=[
                "USE_SOURCE_PERMISSIONS",
                "OPTIONAL",
                "MESSAGE_NEVER",
                "EXCLUDE_FROM_ALL",
                "FILES_MATCHING",
                "EXCLUDE",
            ],
            one_value_keywords=["TYPE", "DESTINATION", "COMPONENT", "PATTERN", "REGEX"],
            multi_value_keywords=[
                "DIRECTORY",
                "FILE_PERMISSIONS",
                "DIRECTORY_PERMISSIONS",
                "CONFIGURATIONS",
                "PERMISSIONS",
            ],
        ),
        "SCRIPT": dict(
            options=["EXCLUDE_FROM_ALL", "ALL_COMPONENTS"],
            one_value_keywords=["SCRIPT", "COMPONENT"],
        ),
        "CODE": dict(
            options=["EXCLUDE_FROM_ALL", "ALL_COMPONENTS"],
            one_value_keywords=["CODE", "COMPONENT"],
        ),
        "EXPORT": dict(
            options=["EXPORT_LINK_INTERFACE_LIBRARIES", "EXCLUDE_FROM_ALL"],
            one_value_keywords=[
                "EXPORT",
                "DESTINATION",
                "NAMESPACE",
                "FILE",
                "COMPONENT",
            ],
            multi_value_keywords=["PERMISSIONS", "CONFIGURATIONS"],
        ),
        "EXPORT_ANDROID_MK": dict(
            options=["EXPORT_LINK_INTERFACE_LIBRARIES", "EXCLUDE_FROM_ALL"],
            one_value_keywords=[
                "EXPORT_ANDROID_MK",
                "DESTINATION",
                "NAMESPACE",
                "FILE",
                "COMPONENT",
            ],
            multi_value_keywords=["PERMISSIONS", "CONFIGURATIONS"],
        ),
        "IMPORTED_RUNTIME_ARTIFACTS": dict(
            options=[
                "LIBRARY",
                "RUNTIME",
                "FRAMEWORK",
                "BUNDLE",
                "OPTIONAL",
                "EXCLUDE_FROM_ALL",
            ],
            one_value_keywords=["RUNTIME_DEPENDENCY_SET", "DESTINATION", "COMPONENT"],
            multi_value_keywords=["PERMISSIONS", "CONFIGURATIONS"],
        ),
        "RUNTIME_DEPENDENCY_SET": dict(
            options=["LIBRARY", "RUNTIME", "FRAMEWORK", "OPTIONAL", "EXCLUDE_FROM_ALL"],
            one_value_keywords=["DESTINATION", "COMPONENT", "NAMELINK_COMPONENT"],
            multi_value_keywords=[
                "PERMISSIONS",
                "CONFIGURATIONS",
                "PRE_INCLUDE_REGEXES",
                "PRE_EXCLUDE_REGEXES",
                "POST_INCLUDE_REGEXES",
                "POST_EXCLUDE_REGEXES",
                "POST_INCLUDE_FILES",
                "POST_EXCLUDE_FILES",
                "DIRECTORIES",
            ],
        ),
    }


class LinkDirectories(ArgumentAwareCommandInvocationDumper):
    options = ["AFTER", "BEFORE"]


class LinkLibraries(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["debug", "optimized", "general"]


class LoadCache(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["pathToBuildDirectory"]
    one_value_keywords = ["READ_WITH_PREFIX"]
    multi_value_keywords = ["EXCLUDE", "INCLUDE_INTERNALS"]


class Project(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<PROJECT-NAME>"]
    one_value_keywords = ["VERSION", "DESCRIPTION", "HOMEPAGE_URL"]
    multi_value_keywords = ["LANGUAGES"]


class SourceGroup(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<name>"]
    one_value_keywords = ["REGULAR_EXPRESSION", "TREE", "PREFIX"]
    multi_value_keywords = ["FILES"]


class TargetCompileDefinitions(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]


class TargetCompileFeatures(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]


class TargetCompileOptions(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    options = ["BEFORE"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]


class TargetIncludeDirectories(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    options = ["BEFORE", "SYSTEM", "AFTER"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]


class TargetLinkDirectories(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    options = ["BEFORE"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]


class TargetLinkOptions(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    options = ["BEFORE"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]


class TargetPrecompileHeaders(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    one_value_keywords = ["REUSE_FROM"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]


class TargetSources(SectionAwareCommandInvocationDumper):
    front_positional_arguments = ["<target>"]
    multi_value_keywords = ["INTERFACE", "PUBLIC", "PRIVATE"]
    sections = {
        "INTERFACE": {
            "one_value_keywords": ["FILE_SET", "TYPE"],
            "multi_value_keywords": ["BASE_DIRS", "FILES"],
        },
        "PUBLIC": {
            "one_value_keywords": ["FILE_SET", "TYPE"],
            "multi_value_keywords": ["BASE_DIRS", "FILES"],
        },
        "PRIVATE": {
            "one_value_keywords": ["FILE_SET", "TYPE"],
            "multi_value_keywords": ["BASE_DIRS", "FILES"],
        },
    }


class TryCompile(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = [
        "<compileResultVar>",
        "<bindir>",
        "<srcdir>",  # or "<srcfile>"
        "<projectName>",
        "<targetName>",
    ]
    option = ["NO_CACHE", "NO_LOG"]
    one_value_keywords = [
        "OUTPUT_VARIABLE",
        "COPY_FILE",
        "COPY_FILE_ERROR",
        "C_STANDARD",
        "C_STANDARD_REQUIRED",
        "C_EXTENSIONS",
        "CXX_STANDARD",
        "CXX_STANDARD_REQUIRED",
        "CXX_EXTENSIONS",
        "OBJC_STANDARD",
        "OBJC_STANDARD_REQUIRED",
        "OBJC_EXTENSIONS",
        "OBJCXX_STANDARD",
        "OBJCXX_STANDARD_REQUIRED",
        "OBJCXX_EXTENSIONS",
        "CUDA_STANDARD",
        "CUDA_STANDARD_REQUIRED",
        "CUDA_EXTENSIONS",
        "PROJECT",
        "SOURCE_DIR",
        "BINARY_DIR",
        "TARGET",
        "LOG_DESCRIPTION",
    ]
    multi_value_keywords = [
        "SOURCES",
        "CMAKE_FLAGS",
        "COMPILE_DEFINITIONS",
        "LINK_OPTIONS",
        "LINK_LIBRARIES",
        "SOURCE_FROM_CONTENT",
        "SOURCE_FROM_VAR",
        "SOURCE_FROM_FILE",
    ]


class TryRun(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = [
        "<runResultVar>",
        "<compileResultVar>",
        "<bindir>",
        "<srcfile>",
    ]
    options = ["NO_CACHE", "NO_LOG"]
    one_value_keywords = [
        "COMPILE_OUTPUT_VARIABLE",
        "RUN_OUTPUT_VARIABLE",
        "OUTPUT_VARIABLE",
        "WORKING_DIRECTORY",
        "COPY_FILE",
        "COPY_FILE_ERROR",
        "C_STANDARD",
        "C_STANDARD_REQUIRED",
        "C_EXTENSIONS",
        "CXX_STANDARD",
        "CXX_STANDARD_REQUIRED",
        "CXX_EXTENSIONS",
        "OBJC_STANDARD",
        "OBJC_STANDARD_REQUIRED",
        "OBJC_EXTENSIONS",
        "OBJCXX_STANDARD",
        "OBJCXX_STANDARD_REQUIRED",
        "OBJCXX_EXTENSIONS",
        "CUDA_STANDARD",
        "CUDA_STANDARD_REQUIRED",
        "CUDA_EXTENSIONS",
        "RUN_OUTPUT_VARIABLE",
        "RUN_OUTPUT_STDOUT_VARIABLE",
        "RUN_OUTPUT_STDERR_VARIABLE",
        "LOG_DESCRIPTION",
    ]
    multi_value_keywords = [
        "CMAKE_FLAGS",
        "COMPILE_DEFINITIONS",
        "LINK_OPTIONS",
        "LINK_LIBRARIES",
        "ARGS",
        "SOURCES",
        "SOURCE_FROM_CONTENT",
        "SOURCE_FROM_VAR",
        "SOURCE_FROM_FILE",
    ]


class SetSourceFilesProperties(
    KeywordWithPairsFormatter, ArgumentAwareCommandInvocationDumper
):
    multi_value_keywords = ["PROPERTIES", "DIRECTORY", "TARGET_DIRECTORY"]
    keyword_formatters = {"PROPERTIES": "_format_keyword_with_pairs"}


class SetTargetProperties(
    KeywordWithPairsFormatter, ArgumentAwareCommandInvocationDumper
):
    multi_value_keywords = ["PROPERTIES"]
    keyword_formatters = {"PROPERTIES": "_format_keyword_with_pairs"}


class SetTestsProperties(
    KeywordWithPairsFormatter, ArgumentAwareCommandInvocationDumper
):
    multi_value_keywords = ["PROPERTIES"]
    keyword_formatters = {"PROPERTIES": "_format_keyword_with_pairs"}


project_command_mapping = {
    "add_custom_command": AddCustomCommand,
    "add_custom_target": AddCustomTarget,
    "add_dependencies": AddDependencies,
    "add_executable": AddExecutable,
    "add_library": AddLibrary,
    "add_subdirectory": AddSubdirectory,
    "add_test": AddTest,
    "build_command": BuildCommand,
    "cmake_file_api": CMakeFileApi,
    "create_test_sourcelist": CreateTestSourcelist,
    "define_property": DefineProperty,
    "export": Export,
    "fltk_wrap_ui": FtlkWrapUi,
    "get_source_file_property": GetSourceFileProperty,
    "get_target_property": GetTargetProperty,
    "get_test_property": GetTestProperty,
    "include_directories": IncludeDirectories,
    "include_external_msproject": IncludeExternalMsProject,
    "install": Install,
    "link_directories": LinkDirectories,
    "link_libraries": LinkLibraries,
    "load_cache": LoadCache,
    "project": Project,
    "source_group": SourceGroup,
    "set_source_files_properties": SetSourceFilesProperties,
    "set_target_properties": SetTargetProperties,
    "set_tests_properties": SetTestsProperties,
    "target_compile_definitions": TargetCompileDefinitions,
    "target_compile_features": TargetCompileFeatures,
    "target_compile_options": TargetCompileOptions,
    "target_include_directories": TargetIncludeDirectories,
    "target_link_directories": TargetLinkDirectories,
    "target_link_libraries": TargetLinkLibraries,
    "target_link_options": TargetLinkOptions,
    "target_precompile_headers": TargetPrecompileHeaders,
    "target_sources": TargetSources,
    "try_compile": TryCompile,
    "try_run": TryRun,
}
